# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import xpart as xp
import xobjects as xo

# Build a reference particle
p0 = xp.Particles(mass0=xp.PROTON_MASS_EV, q0=1, p0c=7e12, x=1, y=3)

# Choose a context
ctx = xo.ContextCpu()

# Built a set of three particles with different y coordinates
particles = xp.build_particles(mode='shift', particle_ref=p0, y=[1,2,3],
                               _context=ctx)

# Inspect
print(particles.p0c[1]) # gives 7e12
print(particles.x[1]) # gives 1.0
print(particles.y[1]) # gives 5.0

#!end-doc-part

assert particles.p0c[1] == 7e12
assert particles.x[1] == 1.0
assert particles.y[1] == 5.0




# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #


import matplotlib.pyplot as plt
import numpy as np

import xobjects as xo


import xpart as xp
import xtrack as xt
import time

ctx = xo.ContextCpu()

# Load machine model
filename = xt._pkg_root.parent.joinpath(
    'test_data/sps_ions/line_and_particle.json')
line = xt.Line.from_json(filename)
line.build_tracker(_context=ctx)

rms_bunch_length=0.14
distribution = "gaussian"
n_particles = 10000
zeta, delta, matcher = xp.generate_longitudinal_coordinates(line.line,
                        num_particles=n_particles,
                        sigma_z=rms_bunch_length, distribution=distribution,
                        engine="single-rf-harmonic", return_matcher=True)
                        #engine="pyheadtail", return_matcher=True)

# Built a set of three particles with different x coordinates
particles = line.build_particles(zeta=0., delta=np.linspace(0,0.0015, 30),
                               x_norm=0, # in sigmas
                               px_norm=0, # in sigmas
                               nemitt_x=3e-6, nemitt_y=3e-6,
                               )

line.track(particles, num_turns=300, turn_by_turn_monitor=True)
z_sep, delta_sep = matcher.get_separatrix()
plt.figure(1)
plt.plot(line.record_last_track.zeta, line.record_last_track.delta, 'k.')
plt.plot(z_sep, delta_sep, 'r')
plt.plot(z_sep, -delta_sep, 'r')
plt.xlim(-3,3)
plt.xlabel('zeta [m]')
plt.ylabel('delta')

plt.show()


# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import json

import matplotlib.pyplot as plt
import numpy as np

import xobjects as xo
import xpart as xp
import xtrack as xt
import time

ctx = xo.ContextCpu()

scenario = 'ions' # can be "protons" or "ions"

if scenario == 'protons':
    filename = xt._pkg_root.parent.joinpath(
        'test_data/sps_w_spacecharge/line_no_spacecharge_and_particle.json')
    with open(filename, 'r') as fid:
        input_data = json.load(fid)
        line = xt.Line.from_dict(input_data['line'])
        line.particle_ref = xp.Particles.from_dict(input_data['particle'])
elif scenario == 'ions':
    # Load machine model (from pymask)
    filename = xt._pkg_root.parent.joinpath(
        'test_data/sps_ions/line_and_particle.json')
    with open(filename, 'r') as fid:
        input_data = json.load(fid)
        line=xt.Line.from_dict(input_data)

line.build_tracker(_context=ctx)


rms_bunch_length=0.25
distribution = "gaussian"
n_particles = 100000
zeta, delta, matcher = xp.generate_longitudinal_coordinates(line=line,
                        num_particles=n_particles,
                        sigma_z=rms_bunch_length, distribution=distribution,
                        engine="single-rf-harmonic", return_matcher=True)
                        #engine="pyheadtail", return_matcher=True)

# Built a set of three particles with different x coordinates
particles = xp.build_particles(_context=ctx,
                               line=line,
                               zeta=zeta, delta=delta,
                               x_norm=0, # in sigmas
                               px_norm=0, # in sigmas
                               scale_with_transverse_norm_emitt=(3e-6, 3e-6)
                               )

zeta_sep, delta_sep = matcher.get_separatrix()
beta0 = line.line.particle_ref.beta0
plt.close('all')
plt.figure(1)
plt.hist2d(zeta, delta, bins=100,
        range=((-1.05 * np.max(zeta_sep), 1.05 * np.max(zeta_sep)),
               (-1.05*np.max(delta_sep), 1.05*np.max(delta_sep))),
        cmin=0.001)
plt.plot(zeta_sep, delta_sep, 'r')
plt.plot(zeta_sep, -delta_sep, 'r')
plt.xlabel('zeta [m]')
plt.ylabel('delta')

zeta_distr_y = matcher.tau_distr_y
zeta_distr_x = matcher.tau_distr_x * beta0
dx = zeta_distr_x[1] - zeta_distr_x[0]
hist, _  = np.histogram(zeta,
                range=(zeta_distr_x[0]-dx/2.,
                zeta_distr_x[-1]+dx/2.), bins=len(zeta_distr_x))
hist = hist / sum(hist) * sum(zeta_distr_y)
plt.figure(2)
plt.plot(zeta_distr_x, hist, 'bo', label="sampled line density")
plt.plot(zeta_distr_x, zeta_distr_y, 'r-', label="input line density")
plt.xlabel('ζ [m]')
plt.legend()

plt.show()


# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import json

import xobjects as xo
import xpart as xp
import xtrack as xt

# Choose a context
ctx = xo.ContextCpu()

# Load machine model (from pymask)
filename = ('../../../xtrack/test_data/lhc_no_bb/line_and_particle.json')
with open(filename, 'r') as fid:
    dct = json.load(fid)
line = xt.Line.from_dict(dct['line'])
line.build_tracker(_context=ctx)

# Attach a reference particle to the line
line.particle_ref = xp.Particles(mass0=xp.PROTON_MASS_EV, q0=1, p0c=7e12, x=1, y=3)

# Built a set of three particles with different x coordinates
particles = line.build_particles(
                               zeta=0, delta=1e-3,
                               x_norm=[1,0,-1], # in sigmas
                               px_norm=[0,1,0], # in sigmas
                               nemitt_x=3e-6, nemitt_y=3e-6)

#!end-doc-part



# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np

import xobjects as xo
import xpart as xp

n_particles = 1000
part = xp.Particles(_capacity=3000,
            px=np.random.uniform(low=-1e-6, high=1e-6, size=n_particles),
            y=np.random.uniform(low=-1e-3, high=1e-3, size=n_particles),
            py=np.random.uniform(low=-1e-6, high=1e-6, size=n_particles),
            zeta=np.random.uniform(low=-1e-2, high=1e-2, size=n_particles),
            delta=np.random.uniform(low=-1e-4, high=1e-4, size=n_particles),
            p0c=7e12)


df_part = part.to_pandas()
df_part_compact = part.to_pandas(compact=True)

assert len(df_part) == 3000
assert len(df_part_compact) == 1000

# Check that underscored vars (if any) are all rng states
for df in [df_part, df_part_compact]:
    ltest = [nn for nn in df.keys() if nn.startswith('_')]
    assert np.all([nn.startswith('_rng') for nn in ltest])

for nn in ['_rng_s1', '_rng_s2', '_rng_s3', '_rng_s4',
            'beta0', 'gamma0', 'ptau', 'rpp', 'rvv']:
    assert nn in df_part.keys()
    assert nn not in df_part_compact.keys()

df_part.to_hdf('part.hdf', key='df', mode='w')
df_part_compact.to_hdf('part_compact.hdf', key='df', mode='w')
df_part.to_hdf('part_compressed.hdf', key='df', mode='w',
               complevel=4, complib='zlib')
df_part_compact.to_hdf('part_compact_compressed.hdf', key='df', mode='w',
                       complevel=4, complib='zlib')

# Measured sizes
# -rw-r--r--  1 giadarol  staff   618K  2 Feb 17:05 part.hdf
# -rw-r--r--  1 giadarol  staff   158K  2 Feb 17:05 part_compact.hdf
# -rw-r--r--  1 giadarol  staff    98K  2 Feb 17:05 part_compressed.hdf
# -rw-r--r--  1 giadarol  staff    74K  2 Feb 17:05 part_compact_compressed.hdf

import pandas as pd
for fname in[
          'part', 'part_compact', 'part_compressed', 'part_compact_compressed']:
    if 'compact' in fname:
        pref = part.remove_unused_space()
    else:
        pref = part
    part_from_pdhdf = xp.Particles.from_pandas(pd.read_hdf(fname +'.hdf'))
    for kk in ['x', 'px', 'y', 'py', 'zeta', 'delta', 'ptau', 'gamma0']:
        assert np.all(getattr(pref, kk) == getattr(part_from_pdhdf, kk))



# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #
import json

import xpart as xp
import xobjects as xo
import xtrack as xt

ctx = xo.ContextCpu() # choose a context

# Load machine model and built a line
filename = ('../../../xtrack/test_data/lhc_no_bb/line_and_particle.json')
with open(filename, 'r') as fid:
    line = xt.Line.from_dict(json.load(fid)['line'])
line.build_tracker(_context=ctx)

# Attach a reference particle to the line
line.particle_ref = xp.Particles(p0c=7e12, mass0=xp.PROTON_MASS_EV, q0=1, x =1 , y=3)

# Built a set of three particles with different y coordinates
# (context and particle_ref are taken from the line)
particles = line.build_particles(y=[1,2,3])

#!end-doc-part

assert particles.p0c[1] == 7e12
assert particles.x[1] == 0.0
assert particles.y[1] == 2.0




# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import xpart as xp

p1 = xp.Particles(x=[1,2,3])
p2 = xp.Particles(x=[4, 5])
p3 = xp.Particles(x=6)

particles = xp.Particles.merge([p1,p2,p3])

print(particles.x) # gives [1. 2. 3. 4. 5. 6.] 


# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

"""
Simple example on how to generate a parabolic particle distribution
"""

import json
import xpart as xp
import xtrack as xt
import xobjects as xo
from xpart import build_particles

from xpart.longitudinal import generate_parabolic_longitudinal_coordinates
import matplotlib.pyplot as plt
import numpy as np

# Decide context
test_on_gpu = False
if test_on_gpu:
    context = xo.ContextCupy()
else:
    context = xo.ContextCpu(omp_num_threads='auto')

# Load the reference particle
filename = ('../../../xtrack/test_data/lhc_no_bb/line_and_particle.json')

with open(filename, 'r') as fid:
    input_data = json.load(fid)
line = xt.Line.from_dict(input_data['line'])
line.build_tracker(_context=context)

# Specify the beam parameters
total_intensity_particles = 1e10
num_particles = 1000000
sigma_z = 0.05
nemitt_x = 3e-6
nemitt_y = 3e-6

# Build a reference particle
p0 = xp.Particles(mass0=xp.PROTON_MASS_EV, q0=1, p0c=7e12, x=1, y=3,
                  delta=[10])

# Build Guassian normalized coordinates
x_norm = np.random.normal(size=num_particles)
px_norm = np.random.normal(size=num_particles)
y_norm = np.random.normal(size=num_particles)
py_norm = np.random.normal(size=num_particles)


# Generate parabolic coordinates
zeta, delta = generate_parabolic_longitudinal_coordinates(num_particles=num_particles,
                                                     	  nemitt_x=nemitt_x, 
                                                     	  nemitt_y=nemitt_y, 
                                                     	  sigma_z=sigma_z,
                                                     	  particle_ref=p0,
                                                     	  line=line)
				
# Build particle object
particles = build_particles(_context=context, particle_ref=p0,
			zeta=zeta, delta=delta, 
			x_norm=x_norm, px_norm=px_norm,
			y_norm=y_norm, py_norm=py_norm,
			nemitt_x=nemitt_x, nemitt_y=nemitt_y,
			weight=total_intensity_particles/num_particles, line=line)

# Make a histogram
bin_heights, bin_borders = np.histogram(particles.zeta, bins=300)
bin_widths = np.diff(bin_borders)
bin_centers = bin_borders[:-1] + bin_widths / 2


# Generate the plots
plt.close('all')

fig1 = plt.figure(1, figsize=(6.4, 7))
ax21 = fig1.add_subplot(3,1,1)
ax22 = fig1.add_subplot(3,1,2)
ax23 = fig1.add_subplot(3,1,3)
ax21.plot(particles.x*1000, particles.px, '.', markersize=1)
ax21.set_xlabel(r'x [mm]')
ax21.set_ylabel(r'px [-]')
ax22.plot(particles.y*1000, particles.py, '.', markersize=1)
ax22.set_xlabel(r'y [mm]')
ax22.set_ylabel(r'py [-]')
ax23.plot(particles.zeta, particles.delta*1000, '.', markersize=1)
ax23.set_xlabel(r'z [-]')
ax23.set_ylabel(r'$\delta$ [1e-3]')
fig1.subplots_adjust(bottom=.08, top=.93, hspace=.33, left=.18,
                     right=.96, wspace=.33)
                     
fig2, ax2 = plt.subplots(1, 1, figsize = (6,5))
ax2.bar(bin_centers, bin_heights, width=bin_widths)
ax2.set_ylabel('Counts')
ax2.set_xlabel(r'z [-]')
                     
plt.show()



import xpart as xp
import numpy as np
import xtrack as xt
import json
import matplotlib.pyplot as plt


filename = xt._pkg_root.joinpath('../../../xsuite/xtrack/test_data/lhc_no_bb/line_and_particle.json')
with open(filename, 'r') as fid:
    input_data = json.load(fid)

line = xt.Line.from_dict(input_data['line'])
line.particle_ref = xp.Particles.from_dict(input_data['particle'])

line.build_tracker()
circumference = line.get_length()
h_list = [35640]
bunch_spacing_in_buckets = 10
bucket_length = circumference/h_list[0]
filling_scheme = np.zeros(int(np.amin(h_list)/bunch_spacing_in_buckets))
n_bunches_tot = 10
filling_scheme[0:int(n_bunches_tot/2)] = 1

filling_scheme[n_bunches_tot:int(3*n_bunches_tot/2)] = 1

bunch_intensity = 1e11
sigma_z = 22.5e-2/5
n_part_per_bunch = int(1e5)
nemitt_x = 2e-6
nemitt_y = 2.5e-6

n_procs = 2
bunche_numbers_per_rank = xp.split_scheme(filling_scheme=filling_scheme,
                                        n_chunk=n_procs)

colors = ['b','g']
for rank in range(n_procs):
    particles = xp.generate_matched_gaussian_multibunch_beam(
             filling_scheme=filling_scheme,
             num_particles=n_part_per_bunch,
             total_intensity_particles=bunch_intensity,
             nemitt_x=nemitt_x, nemitt_y=nemitt_y, sigma_z=sigma_z,
             line=line, bunch_spacing_buckets=10,
             bunch_numbers=bunche_numbers_per_rank[rank],
             particle_ref=line.particle_ref
    )

    print(fr"target sigma_z: {sigma_z}")
    for i_bunch,bunch_number in enumerate(bunche_numbers_per_rank[rank]):
        zeta_rms = np.std(particles.zeta[i_bunch*n_part_per_bunch:
                                         (i_bunch+1)*n_part_per_bunch])
        print(f'zeta rms of bunch {i_bunch}: {zeta_rms}')

    plt.figure(0)
    plt.plot(particles.zeta/bucket_length, particles.delta, 'x',color=colors[rank])
filled_slots = filling_scheme.nonzero()[0]
for filled_slot in filled_slots:
    plt.figure(0)
    plt.axvline(filled_slot*bunch_spacing_in_buckets,color='k',ls='--')

plt.show()



# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import json

import numpy as np
from scipy.constants import e as qe
from scipy.constants import m_p

import xpart as xp
import xtrack as xt

bunch_intensity = 1e11
sigma_z = 25e-2
n_part = int(5e5)
nemitt_x = 2e-6
nemitt_y = 2.5e-6

filename = ('../../../xtrack/test_data/sps_ions/line_and_particle.json')
with open(filename, 'r') as fid:
    ddd = json.load(fid)
line = xt.Line.from_dict(ddd)
line.build_tracker()

particles = xp.generate_matched_gaussian_bunch(
         num_particles=n_part, total_intensity_particles=bunch_intensity,
         nemitt_x=nemitt_x, nemitt_y=nemitt_y, sigma_z=sigma_z,
         engine='single-rf-harmonic',
         line=line)


#!end-doc-part

# CHECKS

y_rms = np.std(particles.y)
py_rms = np.std(particles.py)
x_rms = np.std(particles.x)
px_rms = np.std(particles.px)
delta_rms = np.std(particles.delta)
zeta_rms = np.std(particles.zeta)



import matplotlib.pyplot as plt
plt.close('all')
fig1 = plt.figure(1, figsize=(6.4, 7))
ax21 = fig1.add_subplot(3,1,1)
ax22 = fig1.add_subplot(3,1,2)
ax23 = fig1.add_subplot(3,1,3)
ax21.plot(particles.x*1000, particles.px, '.', markersize=1)
ax21.set_xlabel(r'x [mm]')
ax21.set_ylabel(r'px [-]')
ax22.plot(particles.y*1000, particles.py, '.', markersize=1)
ax22.set_xlabel(r'y [mm]')
ax22.set_ylabel(r'py [-]')
ax23.plot(particles.zeta, particles.delta*1000, '.', markersize=1)
ax23.set_xlabel(r'z [-]')
ax23.set_ylabel(r'$\delta$ [1e-3]')
fig1.subplots_adjust(bottom=.08, top=.93, hspace=.33, left=.18,
                     right=.96, wspace=.33)
plt.show()



# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np

import xobjects as xo
import xpart as xp

# Create a Particles on your selected context (default is CPU)
context = xo.ContextCupy()
part = xp.Particles(_context=context, x=[1,2,3])

# Save particles to json
import json
with open('part.json', 'w') as fid:
    json.dump(part.to_dict(), fid, cls=xo.JEncoder)

# Load particles from json file to selected context
with open('part.json', 'r') as fid:
    part_from_json= xp.Particles.from_dict(json.load(fid), _context=context)




# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np

import xpart as xp
import xtrack as xt

cav = xt.Cavity(frequency=400e6, voltage=6e6)

part = xp.Particles(p0c=6500e9, x=[1,2,3], delta=1e-4)
cav.track(part)

part2 = xp.Particles(**part.to_dict())

assert np.all(part2.ptau == part.ptau)
assert np.all(part2.delta == part.delta)
assert np.all(part2.rpp == part.rpp)
assert np.all(part2.rvv == part.rvv)




# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np

import xpart as xp


x_norm, px_norm, r_points, theta_points = xp.generate_2D_polar_grid(
        r_range=(0.1, 10), nr=8, theta_range=(0, np.pi/2), ntheta=10)

assert len(x_norm) == len(px_norm) == len(r_points) == len(theta_points) == 80

assert np.isclose(np.max(x_norm), 10)
assert np.isclose(np.max(px_norm), 10)
assert np.isclose(np.min(np.sqrt(x_norm**2 + px_norm**2)), 0.1)

assert np.isclose(np.min(theta_points), 0)
assert np.isclose(np.max(theta_points), np.pi/2)
assert np.isclose(np.min(r_points), 0.1)
assert np.isclose(np.max(r_points), 10)

import matplotlib.pyplot as plt
plt.close('all')
plt.plot(x_norm, px_norm, '.')

plt.show()



# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np

import xobjects as xo
import xpart as xp

# Create a Particles on your selected context (default is CPU)
context = xo.ContextCupy()
part = xp.Particles(_context=context, x=[1,2,3])

##############
# PANDAS/HDF #
##############

# Save particles to hdf file via pandas
import pandas as pd
df = part.to_pandas()
df.to_hdf('part.hdf', key='df', mode='w')

# Read particles from hdf file via pandas
part_from_pdhdf = xp.Particles.from_pandas(pd.read_hdf('part.hdf'))



# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import json
import numpy as np

import xpart as xp
import xtrack as xt

num_particles = 10000
nemitt_x = 2.5e-6
nemitt_y = 3e-6

# Load machine model
filename = ('../../../xtrack/test_data/hllhc15_noerrors_nobb/line_and_particle.json')
with open(filename, 'r') as fid:
    input_data = json.load(fid)
line=xt.Line.from_dict(input_data['line'])
line.particle_ref = xp.Particles.from_dict(input_data['particle'])
line.build_tracker()

# Location of the collimator
at_element = 'tcp.d6l7.b1'
at_s = line.line.get_s_position(at_element) + 1.
y_cut = 3e-3 # position of the jaw
pencil_dr_sigmas = 3 # width of the pencil
side = '+' # side of the pencil

# Horizontal plane: generate gaussian distribution in normalized coordinates
x_in_sigmas, px_in_sigmas = xp.generate_2D_gaussian(num_particles)

# Vertical plane: generate pencil beam in absolute coordinates
y_absolute, py_absolute = xp.generate_2D_pencil_with_absolute_cut(num_particles,
                    plane='y', absolute_cut=y_cut, dr_sigmas=pencil_dr_sigmas,
                    side=side, line=line,
                    nemitt_x=nemitt_x, nemitt_y=nemitt_y,
                    at_element=at_element, match_at_s=at_s)

# Longitudinal plane: generate gaussian distribution matched to bucket
zeta, delta = xp.generate_longitudinal_coordinates(
        num_particles=num_particles, distribution='gaussian',
        sigma_z=10e-2, line=line)

# Combine the three planes
# (the normalized coordinates in y/py are blurred in order to preserve the geometric ones)
particles = line.build_particles(nemitt_x=nemitt_x, nemitt_y=nemitt_y,
                y=y_absolute, py=py_absolute,
                x_norm=x_in_sigmas, px_norm=px_in_sigmas,
                zeta=zeta, delta=delta,
                at_element=at_element, match_at_s=at_s)

# Drift to at_s position for checking
drift_to_at_s = xt.Drift(length=at_s-line.get_s_position(at_element))
drift_to_at_s.track(particles)

# Checks and plots

tw_at_s = line.twiss(at_s=at_s)
norm_coords = tw_at_s.get_normalized_coordinates(
                particles, nemitt_x=nemitt_x, nemitt_y=nemitt_y,
                _force_at_element=0)

x_norm = norm_coords.x_norm
px_norm = norm_coords.px_norm
y_norm = norm_coords.y_norm
py_norm = norm_coords.py_norm


import matplotlib.pyplot as plt
plt.close('all')
fig1 = plt.figure(1, figsize=(6.4, 7))
ax1 = fig1.add_subplot(3,2,1)
ax2 = fig1.add_subplot(3,2,3)
ax3 = fig1.add_subplot(3,2,5)
ax1.plot(x_norm, px_norm, '.', markersize=1)
ax1.set_xlabel(r'x [$\sigma$]')
ax1.set_ylabel(r'px [$\sigma$]')
ax1.set_xlim(np.array([-1, 1])*np.max(np.abs(x_norm))*1.1)
ax1.set_ylim(np.array([-1, 1])*np.max(np.abs(px_norm))*1.1)
ax2.plot(y_norm, py_norm, '.', markersize=1)
ax2.set_xlabel(r'y [$\sigma$]')
ax2.set_ylabel(r'py [$\sigma$]')
ax2.set_xlim(np.array([-1, 1])*np.max(np.abs(y_norm))*1.1)
ax2.set_ylim(np.array([-1, 1])*np.max(np.abs(py_norm))*1.1)
ax3.plot(zeta, delta*1000, '.', markersize=1)
ax3.set_xlabel(r'z [m]')
ax3.set_ylabel(r'$\delta$ [1e-3]')

ax21 = fig1.add_subplot(3,2,2)
ax22 = fig1.add_subplot(3,2,4)
ax23 = fig1.add_subplot(3,2,6)
ax21.plot(particles.x*1000, particles.px, '.', markersize=1)
ax21.set_xlabel(r'x [mm]')
ax21.set_ylabel(r'px [-]')
ax21.set_xlim(np.array([-1, 1])*np.max(np.abs(particles.x*1000))*1.1)
ax21.set_ylim(np.array([-1, 1])*np.max(np.abs(particles.px))*1.1)
ax22.plot(particles.y*1000, particles.py, '.', markersize=1)
ax22.axvline(x=y_cut*1000)
ax22.set_xlabel(r'y [mm]')
ax22.set_ylabel(r'py [-]')
ax22.set_xlim(np.array([-1, 1])*np.max(np.abs(particles.y*1000))*1.1)
ax22.set_ylim(np.array([-1, 1])*np.max(np.abs(particles.py))*1.1)
ax23.plot(particles.zeta, particles.delta*1000, '.', markersize=1)
ax23.set_xlabel(r'z [-]')
ax23.set_ylabel(r'$\delta$ [1e-3]')
fig1.subplots_adjust(bottom=.08, top=.93, hspace=.33,
                     right=.96, wspace=.33)
plt.show()



# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import json
import numpy as np

import xpart as xp
import xtrack as xt

num_particles = 10000
nemitt_x = 2.5e-6
nemitt_y = 3e-6

# Load machine model (from pymask)
filename = ('../../../xtrack/test_data/lhc_no_bb/line_and_particle.json')
with open(filename, 'r') as fid:
    input_data = json.load(fid)
line=xt.Line.from_dict(input_data['line'])
line.particle_ref = xp.Particles.from_dict(input_data['particle'])
line.build_tracker()

# Horizontal plane: generate gaussian distribution in normalized coordinates
x_in_sigmas, px_in_sigmas = xp.generate_2D_gaussian(num_particles)

# Vertical plane: generate pencil distribution in normalized coordinates
pencil_cut_sigmas = 6.
pencil_dr_sigmas = 0.7
y_in_sigmas, py_in_sigmas, r_points, theta_points = xp.generate_2D_pencil(
                             num_particles=num_particles,
                             pos_cut_sigmas=pencil_cut_sigmas,
                             dr_sigmas=pencil_dr_sigmas,
                             side='+-')

# Longitudinal plane: generate gaussian distribution matched to bucket 
zeta, delta = xp.generate_longitudinal_coordinates(
        num_particles=num_particles, distribution='gaussian',
        sigma_z=10e-2, line=line)

# Build particles:
#    - scale with given emittances
#    - transform to physical coordinates (using 1-turn matrix)
#    - handle dispersion
#    - center around the closed orbit
particles = line.build_particles(
            zeta=zeta, delta=delta,
            x_norm=x_in_sigmas, px_norm=px_in_sigmas,
            y_norm=y_in_sigmas, py_norm=py_in_sigmas,
            nemitt_x=nemitt_x, nemitt_y=nemitt_y)

# Absolute coordinates can be inspected in particle.x, particles.px, etc.

# Tracking can be done with:
# tracker.track(particles, num_turns=10)

#!end-doc-part

import matplotlib.pyplot as plt
plt.close('all')
fig1 = plt.figure(1, figsize=(6.4, 7))
ax1 = fig1.add_subplot(3,2,1)
ax2 = fig1.add_subplot(3,2,3)
ax3 = fig1.add_subplot(3,2,5)
ax1.plot(x_in_sigmas, px_in_sigmas, '.', markersize=1)
ax1.set_xlabel(r'x [$\sigma$]')
ax1.set_ylabel(r'px [$\sigma$]')
ax1.set_xlim(-7, 7)
ax1.set_ylim(-7, 7)
ax2.plot(y_in_sigmas, py_in_sigmas, '.', markersize=1)
ax2.set_xlabel(r'y [$\sigma$]')
ax2.set_ylabel(r'py [$\sigma$]')
ax2.axvline(x=pencil_cut_sigmas)
ax2.set_xlim(-7, 7)
ax2.set_ylim(-7, 7)
ax3.plot(zeta, delta*1000, '.', markersize=1)
ax3.set_xlabel(r'z [m]')
ax3.set_ylabel(r'$\delta$ [1e-3]')

ax21 = fig1.add_subplot(3,2,2)
ax22 = fig1.add_subplot(3,2,4)
ax23 = fig1.add_subplot(3,2,6)
ax21.plot(particles.x*1000, particles.px, '.', markersize=1)
ax21.set_xlabel(r'x [mm]')
ax21.set_ylabel(r'px [-]')
ax22.plot(particles.y*1000, particles.py, '.', markersize=1)
ax22.set_xlabel(r'y [mm]')
ax22.set_ylabel(r'py [-]')
ax23.plot(particles.zeta, particles.delta*1000, '.', markersize=1)
ax23.set_xlabel(r'z [-]')
ax23.set_ylabel(r'$\delta$ [1e-3]')
fig1.subplots_adjust(bottom=.08, top=.93, hspace=.33,
                     right=.96, wspace=.33)
plt.show()




# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import xpart as xp
import xobjects as xo

ctx = xo.ContextCpu()

particles = xp.Particles(_context=ctx,
        mass0=xp.PROTON_MASS_EV, q0=1, p0c=7e12, # 7 TeV
        x=[1e-3, 0], px=[1e-6, -1e-6], y=[0, 1e-3], py=[2e-6, 0],
        zeta=[1e-2, 2e-2], delta=[0, 1e-4])



# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import pathlib
import json
import numpy as np

import xobjects as xo
import xtrack as xt
import xpart as xp
from scipy.constants import c as clight


fname_line = '../../../xtrack/test_data/hllhc_14/line_and_particle.json'


##############
# Get a line #
##############

with open(fname_line, 'r') as fid:
     input_data = json.load(fid)
line = xt.Line.from_dict(input_data['line'])
line.build_tracker()

particle_ref = xp.Particles.from_dict(input_data['particle'])

zeta, delta = xp.longitudinal.generate_longitudinal_coordinates(
        num_particles=100000, distribution='gaussian',
        sigma_z=10e-2, particle_ref=particle_ref, line=line)

assert np.isclose(np.std(zeta), 10e-2, rtol=1e-2, atol=0)
assert np.isclose(np.std(delta), 1.21e-4, rtol=1e-2, atol=0)



# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import json

import xpart as xp
import xtrack as xt

# Load machine model and build tracker
filename = ('../../../xtrack/test_data/lhc_no_bb/line_and_particle.json')
with open(filename, 'r') as fid:
    input_data = json.load(fid)
line = xt.Line.from_dict(input_data['line'])
line.particle_ref = xp.Particles.from_dict(input_data['particle'])
line.build_tracker()

# Match distribution at a given element
particles = tracker.build_particles(x_norm=[0,1,2], px_norm=[0,0,0], # in sigmas
                   nemitt_x=2.5e-6, nemitt_y=2.5e-6,
                   at_element='ip2')

# Match distribution at a given s position (100m downstream of ip6)
particles = tracker.build_particles(x_norm=[0,1,2], px_norm=[0,0,0], # in sigmas
                   nemitt_x=2.5e-6, nemitt_y=2.5e-6,
                   at_element='ip6',
                   match_at_s=tracker.line.get_s_position('ip6') + 100
                   )





# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np

import xobjects as xo
import xpart as xp

# Create a Particles on your selected context (default is CPU)
context = xo.ContextCupy()
part = xp.Particles(_context=context, x=[1,2,3])

# Save particles to dict 
dct = part.to_dict()

# Load particles from dict 
part_from_dict = xp.Particles.from_dict(dct, _context=context)

#!end-doc-part
dct = part.to_dict()
assert isinstance(dct['delta'], np.ndarray)



# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import xpart as xp
import xobjects as xo

p1 = xp.Particles(x=[1,2,3])

# Make a copy of p1 in the same context
p2 = p1.copy()

# Alter p1
p1.x += 10

# Inspect
print(p1.x) # gives [11. 12. 13.]
print(p2.x) # gives [1. 2. 3.]

# Copy across contexts
ctxgpu = xo.ContextCupy()
p3 = p1.copy(_context=ctxgpu)

# Inspect
print(p3.x[2]) # gives 13



# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np

import xpart as xp
import xobjects as xo

#context = xo.ContextPyopencl()
context = xo.ContextCpu()
ctx2np = context.nparray_from_context_array
particles = xp.Particles(_context=context, p0c=26e9, delta=[1,2,3])

assert ctx2np(particles.delta[2]) == 3
assert np.isclose(ctx2np(particles.rvv[2]), 1.00061, rtol=0, atol=1e-5)
assert np.isclose(ctx2np(particles.rpp[2]), 0.25, rtol=0, atol=1e-10)
assert np.isclose(ctx2np(particles.ptau[2]), 3.001464*particles._xobject.beta0[0],
                  rtol=0, atol=1e-6)

particles.delta[1] = particles.delta[2]

assert particles.delta[2] == particles.delta[1]
assert particles.ptau[2] == particles.ptau[1]
assert particles.rpp[2] == particles.rpp[1]
assert particles.rvv[2] == particles.rvv[1]



# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import json
import numpy as np

import xpart as xp
import xtrack as xt

num_particles = 10000
nemitt_x = 2.5e-6
nemitt_y = 3e-6

# Load machine model (from pymask)
filename = ('../../../xtrack/test_data/lhc_no_bb/line_and_particle.json')
with open(filename, 'r') as fid:
    input_data = json.load(fid)
line = xt.Line.from_dict(input_data['line'])
line.particle_ref = xp.Particles.from_dict(input_data['particle'])
line.build_tracker()

# Horizontal plane: generate cut halo distribution
(x_in_sigmas, px_in_sigmas, r_points, theta_points
    )= xp.generate_2D_uniform_circular_sector(
                                          num_particles=num_particles,
                                          r_range=(0.6, 0.9), # sigmas
                                          theta_range=(0.25*np.pi, 1.75*np.pi))

# Vertical plane: all particles on the closed orbit
y_in_sigmas = 0.
py_in_sigmas = 0.

# Longitudinal plane: all particles off momentum by 1e-3
zeta = 0.
delta = 1e-3

# Build particles:
#    - scale with given emittances
#    - transform to physical coordinates (using 1-turn matrix)
#    - handle dispersion
#    - center around the closed orbit
particles = line.build_particles(
            zeta=zeta, delta=delta,
            x_norm=x_in_sigmas, px_norm=px_in_sigmas,
            y_norm=y_in_sigmas, py_norm=py_in_sigmas,
            nemitt_x=nemitt_x, nemitt_y=nemitt_y)

# Absolute coordinates can be inspected in particle.x, particles.px, etc.

# Tracking can be done with:
# line.track(particles, num_turns=10)

#!end-doc-part

assert (len(x_in_sigmas) == len(px_in_sigmas)
        == len(r_points) == len(theta_points) == 10000)

assert np.isclose(np.max(np.abs(x_in_sigmas)), 0.9, rtol=1e-2)
assert np.isclose(np.max(px_in_sigmas), 0.9, rtol=1e-2)
assert np.isclose(np.min(np.sqrt(x_in_sigmas**2 + px_in_sigmas**2)), 0.6, rtol=1e-2)

assert np.isclose(np.min(theta_points), 0.25*np.pi, rtol=1e-2)
assert np.isclose(np.max(theta_points), 1.75*np.pi, rtol=1e-2)
assert np.isclose(np.min(r_points), 0.6, rtol=1e-2)
assert np.isclose(np.max(r_points), 0.9, rtol=1e-2)

import matplotlib.pyplot as plt
plt.close('all')
fig1 = plt.figure(1, figsize=(6.4, 7))
ax1 = fig1.add_subplot(3,2,1)
ax2 = fig1.add_subplot(3,2,3)
ax3 = fig1.add_subplot(3,2,5)
ax1.plot(x_in_sigmas, px_in_sigmas, '.', markersize=1)
ax1.set_xlabel(r'x [$\sigma$]')
ax1.set_ylabel(r'px [$\sigma$]')
ax1.set_xlim(-1, 1)
ax1.set_ylim(-1, 1)
ax2.plot(y_in_sigmas, py_in_sigmas, '.', markersize=1)
ax2.set_xlabel(r'y [$\sigma$]')
ax2.set_ylabel(r'py [$\sigma$]')
ax2.set_xlim(-1, 1)
ax2.set_ylim(-1, 1)
ax3.plot(zeta, delta*1000, '.', markersize=1)
ax3.set_xlabel(r'z [m]')
ax3.set_ylabel(r'$\delta$ [1e-3]')
ax3.set_xlim(-0.05, 0.05)
ax3.set_ylim(-1.5, 1.5)

ax21 = fig1.add_subplot(3,2,2)
ax22 = fig1.add_subplot(3,2,4)
ax23 = fig1.add_subplot(3,2,6)
ax21.plot(particles.x*1000, particles.px, '.', markersize=1)
ax21.set_xlabel(r'x [mm]')
ax21.set_ylabel(r'px [-]')
ax22.plot(particles.y*1000, particles.py, '.', markersize=1)
ax22.set_xlabel(r'y [mm]')
ax22.set_ylabel(r'py [-]')
ax22.set_xlim(-1, 1)
ax22.set_ylim(-5e-6, 5e-6)
ax23.plot(particles.zeta, particles.delta*1000, '.', markersize=1)
ax23.set_xlabel(r'z [-]')
ax23.set_ylabel(r'$\delta$ [1e-3]')
ax23.set_xlim(-0.05, 0.05)
ax23.set_ylim(-1.5, 1.5)
fig1.subplots_adjust(bottom=.08, top=.93, hspace=.33,
                     right=.96, wspace=.33)
plt.show()
